library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library teaport_v1_00_a;
use teaport_v1_00_a.all;

library unisim ;
use unisim.vcomponents.all ;

entity core is
    port (
        d                   : in std_logic_vector(31 downto 0);
        dout                : out std_logic_vector(31 downto 0);
        slv_reg_write_sel   : in std_logic_vector(0 to 3);
        slv_reg_read_sel    : in std_logic_vector(0 to 3);
        slv_read_ack        : out std_logic;
        slv_write_ack       : out std_logic;
        reset               : in std_logic;
        sysclock            : in std_logic;
        intr                : out std_logic;
        sout                : out std_logic;
        sin                 : in std_logic);
end entity core;

architecture imp of core is

    constant LEFT_BIT           : Natural := 9;
    constant FULL_THRESHOLD     : Natural := 1020 ;
    constant VERSION            : std_logic_vector ( 7 downto 0 ) := x"13" ;
    constant LOWER              : std_logic_vector ( 7 downto 0 ) := x"40" ;
    constant UPPER              : std_logic_vector ( 7 downto 0 ) := x"7d" ;
    constant NO_IRQ_TIMER_MAX   : Natural := 8388607; -- ~1/12 second.

        
    type StateType is ( BUS_READY, BUS_READ_ACK, BUS_WRITE_ACK,
            BUS_READ_0, BUS_READ_1, BUS_READ_2, TEST_STATE ,
            BUS_READ_3, BUS_READ_4, BUS_READ_5 ) ;

    type ReadStateType is ( NOT_READY_0, NOT_READY_1, READ_READY ) ;

    type MemSourceType is ( SRC_UART, SRC_BUS, SRC_NO_WRITE, SRC_TEST ) ;
    

    signal state                : StateType ;
    signal read_state           : ReadStateType ;
    signal rx_fifo_consumer     : std_logic_vector(LEFT_BIT downto 0);
    signal rx_fifo_producer     : std_logic_vector(LEFT_BIT downto 0);
    signal tx_fifo_consumer     : std_logic_vector(LEFT_BIT downto 0);
    signal tx_fifo_producer     : std_logic_vector(LEFT_BIT downto 0);
    signal rx_fifo_consumer_next: std_logic_vector(LEFT_BIT downto 0);
    signal rx_fifo_producer_next: std_logic_vector(LEFT_BIT downto 0);
    signal tx_fifo_consumer_next: std_logic_vector(LEFT_BIT downto 0);
    signal tx_fifo_producer_next: std_logic_vector(LEFT_BIT downto 0);
    signal rx_count             : std_logic_vector(LEFT_BIT downto 0);
    signal tx_count             : std_logic_vector(LEFT_BIT downto 0);
    signal rx_threshold         : std_logic_vector(LEFT_BIT downto 0);
    signal enable_tx_int        : std_logic;
    signal enable_rx_int        : std_logic;
    signal mem_write            : std_logic;
    signal mem_address          : std_logic_vector(LEFT_BIT+1 downto 0);
    signal baud_divisor         : std_logic_vector(15 downto 0);
    signal ua_rd, ua_wr         : std_logic;
    signal ua_dout, wr_data     : std_logic_vector(7 downto 0);
    signal mem_data_in          : std_logic_vector(7 downto 0);
    signal mem_data_out         : std_logic_vector(7 downto 0);
    signal rf_empty             : std_logic;
    signal tx_buffer_full       : std_logic;
    signal rx_overflow          : std_logic;
    signal tx_overflow          : std_logic;
    signal internal_overflow    : std_logic;
    signal internal_overflow_sig: std_logic;
    signal ua_reset             : std_logic;

    signal sout_i, sin_i        : std_logic;

    signal no_irq_timer         : Natural range 0 to NO_IRQ_TIMER_MAX;
    signal reset_no_irq_timer   : std_logic;

    signal test_mode            : std_logic;
    signal test_data            : std_logic_vector(7 downto 0);
    signal mem_source           : MemSourceType;

    signal rx_empty             : boolean ;
    signal tx_empty             : boolean ;
    signal rx_full              : boolean ;
    signal tx_full              : boolean ;
    signal rx_trigger           : boolean ;

begin


    controller : process ( sysclock , reset ) is
    begin
        if ( reset = '1' )
        then
            state <= BUS_READY;
            enable_tx_int <= '0';
            enable_rx_int <= '0';
            baud_divisor <= x"028b" ; -- 9600
            ua_reset <= '1';
            reset_no_irq_timer <= '1';
            test_data <= LOWER ;
            test_mode <= '0';
            rx_threshold <= ( others => '0' );
            rx_threshold ( 4 ) <= '1' ;
            rx_threshold ( 3 downto 0 ) <= ( others => '0' );

        elsif ( sysclock'event )
        and ( sysclock = '1' )
        then
            slv_read_ack <= '0';
            slv_write_ack <= '0';
            ua_rd <= '0';
            ua_wr <= '0';
            mem_source <= SRC_NO_WRITE;
            internal_overflow <= internal_overflow_sig or internal_overflow ; 
            read_state <= NOT_READY_0 ;
            reset_no_irq_timer <= '0';

            if ( ua_reset = '1' )
            then
                rx_fifo_consumer <= ( others => '0' ) ;
                rx_fifo_producer <= ( others => '0' ) ;
                tx_fifo_consumer <= ( others => '0' ) ;
                tx_fifo_producer <= ( others => '0' ) ;
            end if ;

            case state is
            when BUS_READY =>
                dout ( 27 downto 0 ) <= ( others => '0' ) ;
                dout ( 31 ) <= rx_overflow ;
                dout ( 30 ) <= tx_overflow ;
                dout ( 29 ) <= internal_overflow ;
                dout ( 28 ) <= ua_reset ;
                if ( slv_reg_read_sel /= "0000" )
                then
                    state <= BUS_READ_ACK;
                    if ( slv_reg_read_sel ( 0 ) = '1' )
                    then
                        -- Read data register - send byte(s) to bus
                        if ( rx_empty )
                        then
                            -- RX FIFO empty
                            null;
                        else
                            -- RX FIFO has some data in it
                            state <= BUS_READ_0;
                            mem_address <= rx_fifo_consumer & "0" ;
                            rx_fifo_consumer <= rx_fifo_consumer_next;
                        end if;
                        -- Simply reading this register is enough
                        -- to reset the IRQ timer
                        reset_no_irq_timer <= '1';
                    elsif ( slv_reg_read_sel ( 1 ) = '1' )
                    then
                        -- Read status register
                        if ( not tx_full )
                        then
                            -- TX FIFO not full
                            dout ( 0 ) <= '1';
                        end if;
                        if ( tx_empty )
                        then
                            -- TX FIFO empty
                            dout ( 1 ) <= '1';
                        end if;
                        if ( not rx_empty )
                        then
                            -- RX FIFO not empty
                            dout ( 2 ) <= '1';
                        end if;
                    elsif ( slv_reg_read_sel ( 2 ) = '1' )
                    then
                        dout ( 7 downto 0 ) <= VERSION ;
                    end if;
                elsif ( slv_reg_write_sel /= "0000" )
                then
                    -- Received byte from bus
                    state <= BUS_WRITE_ACK;
                    if ( slv_reg_write_sel ( 0 ) = '1' )
                    then
                        -- Write data register - put byte in FIFO
                        if ( tx_full )
                        then
                            -- TX FIFO full
                            tx_overflow <= '1';
                        else
                            mem_source <= SRC_BUS;
                            mem_address <= tx_fifo_producer & "1" ;
                            tx_fifo_producer <= tx_fifo_producer_next ;
                        end if;
                    elsif ( slv_reg_write_sel ( 1 ) = '1' )
                    then
                        -- Write status register - turn interrupts on/off
                        if ( d ( 0 ) = '1' )
                        then
                            enable_rx_int <= '0';
                        elsif ( d ( 1 ) = '1' )
                        then
                            enable_rx_int <= '1';
                        end if;
                        if ( d ( 2 ) = '1' )
                        then
                            enable_tx_int <= '0';
                        elsif ( d ( 3 ) = '1' )
                        then
                            enable_tx_int <= '1';
                        end if;
                        -- Reset control; clears the FIFO and
                        -- resets the Martin Schoeberl UART.
                        ua_reset <= '0' ;
                        if ( d ( 4 ) = '1' )
                        then
                            ua_reset <= '1' ;
                            test_mode <= '0' ;
                        end if;
                        if ( d ( 5 ) = '1' )
                        then
                            -- Clear error flags
                            rx_overflow <= '0' ;
                            tx_overflow <= '0' ;
                            internal_overflow <= '0' ;
                        end if;
                        if ( d ( 6 ) = '1' )
                        then
                            test_mode <= '1';
                            test_data <= LOWER ;
                        end if;
                    elsif ( slv_reg_write_sel ( 2 ) = '1' )
                    then
                        -- Write baud register
                        baud_divisor <= d ( 15 downto 0 );
                        rx_threshold <= d ( 16 + LEFT_BIT downto 16 ) ;
                        rx_threshold ( 3 downto 0 ) <= ( others => '0' );
                    end if;
                elsif (( rf_empty = '0' )
                and ( ua_rd = '0' )
                and ( ua_wr = '0' ))
                then
                    -- Received byte from UART
                    ua_rd <= '1';
                    if ( rx_full )
                    then
                        -- RX FIFO full, don't write
                        rx_overflow <= '1';
                    else
                        rx_fifo_producer <= rx_fifo_producer_next;
                        mem_source <= SRC_UART;
                        mem_address <= rx_fifo_producer & "0" ;
                    end if;
                elsif (( not tx_empty )
                and ( tx_buffer_full = '0' ))
                then
                    -- Prepare to send byte through UART
                    -- It takes time to fetch the byte from the BRAM.
                    mem_address <= tx_fifo_consumer & "1" ;
                    case read_state is
                    when NOT_READY_0 =>
                        read_state <= NOT_READY_1 ;
                    when NOT_READY_1 =>
                        read_state <= READ_READY ;
                    when READ_READY =>
                        -- Actually send byte through UART
                        tx_fifo_consumer <= tx_fifo_consumer_next;
                        ua_wr <= '1';
                        read_state <= NOT_READY_0 ;
                    end case;
                elsif (( test_mode = '1' )
                and ( tx_empty ))
                then
                    -- TX FIFO empty in test mode. Add dummy bytes.
                    -- Add dummy bytes to TX FIFO.
                    mem_source <= SRC_TEST;
                    mem_address <= tx_fifo_producer & "1" ;
                    tx_fifo_producer <= tx_fifo_producer_next ;
                    state <= TEST_STATE;
                end if;

            when TEST_STATE =>
                if ( test_data = UPPER )
                then
                    test_data <= LOWER ;
                else
                    test_data <= test_data + 1 ;
                end if;
                state <= BUS_READY;

            when BUS_WRITE_ACK =>
                if ( slv_reg_write_sel /= "0000" ) 
                then
                    slv_write_ack <= '1';
                else
                    state <= BUS_READY;
                end if;

            when BUS_READ_ACK =>
                if ( slv_reg_read_sel /= "0000" )
                then
                    slv_read_ack <= '1';
                else
                    state <= BUS_READY;
                end if;

            when BUS_READ_0 =>
                -- It takes time to fetch the byte from the BRAM.
                state <= BUS_READ_1 ;

            when BUS_READ_1 =>
                dout ( 24 ) <= '1' ;
                dout ( 7 downto 0 ) <= mem_data_out;
                if ( rx_empty )
                then
                    state <= BUS_READ_ACK;
                else
                    state <= BUS_READ_2;
                    mem_address <= rx_fifo_consumer & "0" ;
                    rx_fifo_consumer <= rx_fifo_consumer_next;
                end if;

            when BUS_READ_2 =>
                state <= BUS_READ_3;

            when BUS_READ_3 =>
                dout ( 25 ) <= '1' ;
                dout ( 15 downto 8 ) <= mem_data_out;
                if ( rx_empty )
                then
                    state <= BUS_READ_ACK;
                else
                    state <= BUS_READ_4;
                    mem_address <= rx_fifo_consumer & "0" ;
                    rx_fifo_consumer <= rx_fifo_consumer_next;
                end if;

            when BUS_READ_4 =>
                state <= BUS_READ_5;

            when BUS_READ_5 =>
                dout ( 26 ) <= '1' ;
                dout ( 23 downto 16 ) <= mem_data_out;
                if ( rx_empty )
                then
                    -- FIFO now empty
                    state <= BUS_READ_ACK;
                else
                    -- FIFO still contains data
                    dout ( 27 ) <= '1';
                    state <= BUS_READ_ACK;
                end if;
            end case ;
        end if;
    end process controller;

    -- FIFO memory
    -- Even addresses: RX FIFO
    -- Odd addresses: TX FIFO
    mem_data_in <= ua_dout when ( mem_source = SRC_UART ) 
               else d ( 7 downto 0 ) when ( mem_source = SRC_BUS )
               else test_data;
    mem_write <= '0' when mem_source = SRC_NO_WRITE else '1';
    
    overflow_fifo : RAMB16_S9
        port map ( 
                addr => mem_address,
                di => mem_data_in,
                dip => "0",
                do => mem_data_out,
                dop => open,
                clk => sysclock,
                en => '1',
                ssr => '0',
                we => mem_write ) ;
                   

    -- Producer/consumer counters
    rx_fifo_consumer_next <= rx_fifo_consumer + 1;
    rx_fifo_producer_next <= rx_fifo_producer + 1;
    tx_fifo_consumer_next <= tx_fifo_consumer + 1;
    tx_fifo_producer_next <= tx_fifo_producer + 1;
    rx_count <= rx_fifo_producer - rx_fifo_consumer ;
    tx_count <= tx_fifo_producer - tx_fifo_consumer ;

    -- empty/full
    rx_empty <= ( rx_count = 0 ) ;
    tx_empty <= ( tx_count = 0 ) ;
    rx_full <= ( rx_count >= FULL_THRESHOLD ) ;
    tx_full <= ( tx_count >= FULL_THRESHOLD ) ;
    rx_trigger <= ( rx_count >= rx_threshold ) ;

    -- Interrupts
    process ( sysclock ) is
    begin
        if ( reset = '1' )
        then
            intr <= '0' ;
            no_irq_timer <= 0 ;

        elsif ( sysclock'event )
        and ( sysclock = '1' )
        then
            intr <= '0' ;
            if (( enable_rx_int = '1' ) 
            and ( not rx_empty )) 
            then
                if ( rx_trigger )
                then
                    -- Triggered because RX FIFO is getting full
                    intr <= '1' ;
                elsif ( no_irq_timer = NO_IRQ_TIMER_MAX )
                then
                    -- Triggered because RX FIFO is not empty and
                    -- a timeout has been reached
                    intr <= '1' ;
                else
                    no_irq_timer <= no_irq_timer + 1 ;
                end if;
            end if;

            if (( enable_tx_int = '1' )
            and ( tx_empty ))
            then
                -- Triggered because the TX FIFO is empty.
                intr <= '1' ;
            end if;

            if ( reset_no_irq_timer = '1' )
            then
                no_irq_timer <= 0;
            end if;
        end if;
    end process;

    -- Martin Schoeberl's GPLv3 UART
    ms_uart : entity teaport_v1_00_a.sc_uart 
    port map (
        clk => sysclock,
        reset => ua_reset,
        ua_dout => ua_dout,
        ua_wr => ua_wr,
        ua_rd => ua_rd,
        wr_data	=> mem_data_out,
        rf_empty => rf_empty,
        tx_buffer_full => tx_buffer_full,

        rx_buffer_overflow => internal_overflow_sig,
        clk16_cnt => baud_divisor,

        txd => sout_i,
        rxd => sin_i );

    -- In test mode, sin is looped to sout.
    sout <= sout_i;
    sin_i <= sin when test_mode = '0' else sout_i;

end architecture imp;

